@@ -1,4 +1,5 @@
 import pandas as pd
+from pandas.api.types import is_complex_dtype
 import numpy as np
 import gemmi
 import reciprocalspaceship as rs
@@ -8,11 +9,51 @@ from reciprocalspaceship.utils import (
     apply_to_hkl,
     phase_shift,
     is_centric,
+    is_absent,
     in_asu,
     hkl_to_asu,
     hkl_to_observed,
     compute_dHKL,
+    compute_structurefactor_multiplicity,
+    bin_by_percentile
 )
+from functools import wraps
+from inspect import signature
+
+
+def inplace(f):
+    """ 
+    A decorator that applies the inplace argument. Base function must have a Bool param called "inplace".
+    The position of `inplace` doesn't matter.
+    """
+    @wraps(f)
+    def wrapped(ds, *args, **kwargs):
+        sig = signature(f)
+        bargs = sig.bind(ds, *args, **kwargs)
+        bargs.apply_defaults()
+        if 'inplace' in bargs.arguments:
+            if bargs.arguments['inplace']:
+                return f(ds, *args, **kwargs)
+            else:
+                return f(ds.copy(), *args, **kwargs)
+        else:
+            raise KeyError(f'"inplace" not found in local variables of @inplacemethod decorated function {f} '
+                             "Edit your method definition to include `inplace=Bool`. "
+                             )
+    return wrapped
+
+def range_indexed(f):
+    """ A decorator that presents the dataset with a range index and makes sure the method preserves the original index columns if any """
+    @wraps(f)
+    def wrapped(ds, *args, **kwargs):
+        names = ds.index.names
+        ds = ds._index_from_names([None], inplace=True)
+        result = f(ds, *args, **kwargs)
+        result = ds._index_from_names(names, inplace=True)
+        if result is not ds:
+            ds = ds._index_from_names(names, inplace=True)
+        return result.__finalize__(ds)
+    return wrapped
 
 class DataSet(pd.DataFrame):
     """
@@ -78,8 +119,14 @@ class DataSet(pd.DataFrame):
 
     @spacegroup.setter
     def spacegroup(self, val):
-        self._spacegroup = val
-    
+        # GH#18: Type-checking for supported input types
+        if isinstance(val, gemmi.SpaceGroup) or (val is None):
+            self._spacegroup = val
+        elif isinstance(val, (str, int)):
+            self._spacegroup = gemmi.SpaceGroup(val)
+        else:
+            raise ValueError(f"Cannot construct gemmi.SpaceGroup from value: {val}")
+            
     @property
     def cell(self):
         """Unit cell parameters (a, b, c, alpha, beta, gamma)"""
@@ -87,8 +134,15 @@ class DataSet(pd.DataFrame):
 
     @cell.setter
     def cell(self, val):
-        self._cell = val
-
+        # GH#18: Type-checking for supported input types
+        if isinstance(val, gemmi.UnitCell) or (val is None):
+            self._cell = val
+        elif isinstance(val, (list, tuple, np.ndarray)) and len(val) == 6:
+            self._cell = gemmi.UnitCell(*val)
+        else:
+            raise ValueError(f"Cannot construct gemmi.UnitCell from value: {val}")
+            
+            
     @property
     def merged(self):
         """Whether DataSet contains merged reflection data (boolean)"""
@@ -100,7 +154,19 @@ class DataSet(pd.DataFrame):
 
     #-------------------------------------------------------------------
     # Methods
-    
+
+    @inplace
+    def _index_from_names(self, names, inplace=False):
+        """ Helper method for dectorators """
+        if names == [None] and self.index.names != [None]:
+            self.reset_index(inplace=True)
+        elif names != [None] and self.index.names == [None]:
+            self.set_index(names, inplace=True)
+        elif names != self.index.names:
+            self.reset_index(inplace=True)
+            self.set_index(names, inplace=True)
+        return self
+
     def set_index(self, keys, drop=True, append=False, inplace=False, verify_integrity=False):
         """
         Set the DataSet index using existing columns.
@@ -258,7 +324,53 @@ class DataSet(pd.DataFrame):
         """
         from reciprocalspaceship import io
         return io.to_gemmi(self, skip_problem_mtztypes)
-    
+
+    def to_structurefactor(self, sf_key, phase_key):
+        """
+        Convert structure factor amplitudes and phases to complex structure
+        factors
+
+        Parameters
+        ----------
+        sf_key : str
+            Column label for structure factor amplitudes
+        phase_key : str
+            Column label for phases
+
+        Returns
+        -------
+        rs.DataSeries
+            Complex structure factors
+
+        See Also
+        --------
+        DataSet.from_structurefactor : Convert complex structure factor to amplitude and phase
+        """
+        sfs = utils.to_structurefactor(self[sf_key], self[phase_key])
+        return rs.DataSeries(sfs, index=self.index)
+
+    def from_structurefactor(self, sf_key):
+        """
+        Convert complex structure factors to structure factor amplitudes
+        and phases
+
+        Parameters
+        ----------
+        sf_key : str
+            Column label for complex structure factors
+        
+        Returns
+        -------
+        (sf, phase) : tuple of DataSeris
+            Tuple of DataSeries for the structure factor amplitudes and
+            phases corresponding to the complex structure factors
+
+        See Also
+        --------
+        DataSet.to_structurefactor : Convert amplitude and phase to complex structure factor
+        """
+        return utils.from_structurefactor(self[sf_key])
+        
     def append(self, *args, check_isomorphous=True, **kwargs):
         """
         Append rows of `other` to the end of calling DataSet, returning
@@ -403,6 +515,18 @@ class DataSet(pd.DataFrame):
         keys = [ k for k in self if isinstance(self.dtypes[k], rs.PhaseDtype) ]
         return keys
 
+    def get_complex_keys(self):
+        """
+        Return columns labels for data with complex dtype.
+
+        Returns
+        -------
+        keys : list of strings
+            list of column labels with complex dtype
+        """
+        keys = [ k for k in self if is_complex_dtype(self.dtypes[k]) ]
+        return keys
+
     def get_m_isym_keys(self):
         """
         Return column labels for data with M/ISYM dtype.
@@ -414,7 +538,9 @@ class DataSet(pd.DataFrame):
         """
         keys = [ k for k in self if isinstance(self.dtypes[k], rs.M_IsymDtype) ]
         return keys
-    
+
+    @inplace
+    @range_indexed
     def apply_symop(self, symop, inplace=False):
         """
         Apply symmetry operation to all reflections in DataSet object. 
@@ -431,27 +557,33 @@ class DataSet(pd.DataFrame):
         elif not isinstance(symop, gemmi.Op):
             raise ValueError(f"Provided symop is not of type gemmi.Op")
 
-        if inplace:
-            F = self
+        dataset = self
+
+        # Handle phase flips associated with Friedel operator
+        if symop.det_rot() < 0:
+            phic = -1
         else:
-            F = self.copy()
-            
+            phic = 1
+
         # Apply symop to generate new HKL indices and phase shifts
-        H = F.get_hkls()
+        H = dataset.get_hkls()
         hkl = apply_to_hkl(H, symop)
-        phase_shifts = np.rad2deg(phase_shift(H, symop))
+        phase_shifts = phase_shift(H, symop)
             
-        F.reset_index(inplace=True)
-        F[['H', 'K', 'L']] = hkl
-        F[['H', 'K', 'L']] = F[['H', 'K', 'L']].astype(rs.HKLIndexDtype())
-        F.set_index(['H', 'K', 'L'], inplace=True)
+        dataset[['H', 'K', 'L']] = hkl
+        dataset[['H', 'K', 'L']] = dataset[['H', 'K', 'L']].astype(rs.HKLIndexDtype())
 
         # Shift phases according to symop
-        for key in F.get_phase_keys():
-            F[key] += phase_shifts
-            F[key] = utils.canonicalize_phases(F[key], deg=True)
-            
-        return F.__finalize__(self)
+        for key in dataset.get_phase_keys():
+            dataset[key] += np.rad2deg(phase_shifts)
+            dataset[key] *= phic
+            dataset[key] = utils.canonicalize_phases(dataset[key], deg=True)
+        for key in dataset.get_complex_keys():
+            dataset[key] *= np.exp(1j*phase_shifts)
+            if symop.det_rot() < 0:
+                dataset[key] = np.conjugate(dataset[key])
+        
+        return dataset
 
     def get_hkls(self):
         """
@@ -465,6 +597,7 @@ class DataSet(pd.DataFrame):
         hkl = self.reset_index()[['H', 'K', 'L']].to_numpy(dtype=np.int32)
         return hkl
 
+    @inplace
     def label_centrics(self, inplace=False):
         """
         Label centric reflections in DataSet. A new column of
@@ -475,14 +608,24 @@ class DataSet(pd.DataFrame):
         inplace : bool
             Whether to add the column in place or to return a copy
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
+        self['CENTRIC'] = is_centric(self.get_hkls(), self.spacegroup)
+        return self
 
-        dataset['CENTRIC'] = is_centric(dataset.get_hkls(), dataset.spacegroup)
-        return dataset
+    @inplace
+    def label_absences(self, inplace=False):
+        """
+        Label systematically absent reflections in DataSet. A new column 
+        of booleans, "ABSENT", is added to the object.
 
+        Parameters
+        ----------
+        inplace : bool
+            Whether to add the column in place or to return a copy
+        """
+        self['ABSENT'] = is_absent(self.get_hkls(), self.spacegroup)
+        return self
+
+    @inplace
     def infer_mtz_dtypes(self, inplace=False, index=True):
         """
         Infers MTZ dtypes from column names and underlying data. This 
@@ -508,25 +651,21 @@ class DataSet(pd.DataFrame):
         --------
         DataSeries.infer_mtz_dtype : Infer MTZ dtype for DataSeries
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
-
         # See GH#2: Handle unnamed Index objects such as RangeIndex
         if index:
-            index_keys = list(filter(None, dataset.index.names))
+            index_keys = list(filter(None, self.index.names))
             if index_keys:
-                dataset.reset_index(inplace=True, level=index_keys)
+                self.reset_index(inplace=True, level=index_keys)
 
-        for c in dataset:
-            dataset[c] = dataset[c].infer_mtz_dtype()
+        for c in self:
+            self[c] = self[c].infer_mtz_dtype()
 
         if index and index_keys:
-            dataset.set_index(index_keys, inplace=True)
+            self.set_index(index_keys, inplace=True)
 
-        return dataset
+        return self
 
+    @inplace
     def compute_dHKL(self, inplace=False):
         """
         Compute the real space lattice plane spacing, d, associated with
@@ -537,15 +676,29 @@ class DataSet(pd.DataFrame):
         inplace : bool
             Whether to add the column in place or return a copy
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
-
         dHKL = compute_dHKL(self.get_hkls(), self.cell)
-        dataset['dHKL'] = rs.DataSeries(dHKL, dtype='R', index=dataset.index)
-        return dataset
+        self['dHKL'] = rs.DataSeries(dHKL, dtype='R', index=self.index)
+        return self
 
+    @inplace
+    def compute_multiplicity(self, inplace=False, include_centering=True):
+        """
+        Compute the multiplicity of reflections in DataSet. A new column of
+        floats, "EPSILON", is added to the object.
+
+        Parameters
+        ----------
+        inplace : bool
+            Whether to add the column in place or to return a copy
+        include_centering : bool
+            Whether to include centering operations in the multiplicity calculation.
+            The default is to include them.
+        """
+        epsilon = compute_structurefactor_multiplicity(self.get_hkls(), self.spacegroup, include_centering)
+        self['EPSILON'] = rs.DataSeries(epsilon, dtype='I', index=self.index)
+        return self
+
+    @inplace
     def assign_resolution_bins(self, bins=20, inplace=False, return_labels=True):
         """
         Assign reflections in DataSet to resolution bins.
@@ -564,19 +717,15 @@ class DataSet(pd.DataFrame):
         -------
         (DataSet, list) or DataSet
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
-        dHKL = dataset.compute_dHKL()["dHKL"]
+        dHKL = self.compute_dHKL()["dHKL"]
 
-        assignments, labels = rs.utils.bin_by_percentile(dHKL, bins=bins, ascending=False)
-        dataset["bin"] = rs.DataSeries(assignments, dtype="I", index=dataset.index)
+        assignments, labels = bin_by_percentile(dHKL, bins=bins, ascending=False)
+        self["bin"] = rs.DataSeries(assignments, dtype="I", index=self.index)
 
         if return_labels:
-            return dataset, labels
+            return self, labels
         else:
-            return dataset
+            return self
         
     def stack_anomalous(self, plus_labels=None, minus_labels=None):
         """
@@ -799,6 +948,8 @@ class DataSet(pd.DataFrame):
         
         return True
 
+    @inplace
+    @range_indexed
     def hkl_to_asu(self, inplace=False):
         """
         Map HKL indices to the reciprocal space asymmetric unit. If phases
@@ -823,15 +974,7 @@ class DataSet(pd.DataFrame):
         --------
         DataSet.hkl_to_observed : Opposite of DataSet.hkl_to_asu()
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
-
-        index_keys = None
-        if "H" in dataset.index.names or "K" in dataset.index.names or "L" in dataset.index.names:
-            index_keys = dataset.index.names
-            dataset.reset_index(inplace=True)
+        dataset = self
 
         # Compute new HKLs and phase shifts
         hkls = dataset.get_hkls()
@@ -846,9 +989,6 @@ class DataSet(pd.DataFrame):
         dataset[["H", "K", "L"]] = asu_hkls[inverse]
         dataset[["H", "K", "L"]] = dataset[["H", "K", "L"]].astype("HKL")
 
-        if index_keys is not None:
-            dataset.set_index(index_keys, inplace=True)
-
         # Apply phase shift
         for k in dataset.get_phase_keys():
             dataset[k] = phi_coeff[inverse] * (dataset[k] + phi_shift[inverse])
@@ -864,6 +1004,8 @@ class DataSet(pd.DataFrame):
 
         return dataset
 
+    @inplace
+    @range_indexed
     def hkl_to_observed(self, m_isym=None, inplace=False):
         """
         Map HKL indices to their observed index using an ``M/ISYM`` column. 
@@ -895,52 +1037,78 @@ class DataSet(pd.DataFrame):
         --------
         DataSet.hkl_to_asu : Opposite of DataSet.hkl_to_observed()
         """ 
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
-
         # Validate input
         if m_isym is None:
-            m_isym = dataset.get_m_isym_keys()
+            m_isym = self.get_m_isym_keys()
             if len(m_isym) == 1:
                 m_isym = m_isym[0]
             else:
                 raise ValueError(f"Method requires a single M/ISYM column -- found: {m_isym}")
         elif not isinstance(m_isym, str):
             raise ValueError("Provided M/ISYM column label should be type str")
-        elif not isinstance(dataset.dtypes[m_isym], rs.M_IsymDtype):
+        elif not isinstance(self.dtypes[m_isym], rs.M_IsymDtype):
             raise ValueError(f"Provided M/ISYM column label is of wrong dtype")
 
         # GH#3: Separate combined M/ISYM into M and ISYM        
-        isym = (dataset[m_isym] % 256).to_numpy(dtype=np.int32)
-        if not dataset.merged:
-            dataset["PARTIAL"] = (dataset[m_isym]/256).astype(int) != 0
-        dataset.drop(columns=m_isym, inplace=True)
+        isym = (self[m_isym] % 256).to_numpy(dtype=np.int32)
+        if not self.merged:
+            self["PARTIAL"] = (self[m_isym]/256).astype(int) != 0
+        self.drop(columns=m_isym, inplace=True)
         
         # Compute new HKLs and phase shifts
-        hkls = dataset.get_hkls()
+        hkls = self.get_hkls()
         hkls_isym = np.concatenate([hkls, isym.reshape(-1, 1)], axis=1)
         compressed, inverse = np.unique(hkls_isym, axis=0, return_inverse=True)
         observed_hkls, phi_coeff, phi_shift = hkl_to_observed(
             compressed[:, :3],       # compressed HKLs
             compressed[:, 3],        # compressed ISYM
-            dataset.spacegroup,
+            self.spacegroup,
             return_phase_shifts=True
         )
-        index_keys = dataset.index.names
-        dataset.reset_index(inplace=True)
-        dataset[["H", "K", "L"]] = observed_hkls[inverse]
-        dataset[["H", "K", "L"]] = dataset[["H", "K", "L"]].astype("HKL")
-        dataset.set_index(index_keys, inplace=True)
+        self[["H", "K", "L"]] = observed_hkls[inverse]
+        self[["H", "K", "L"]] = self[["H", "K", "L"]].astype("HKL")
 
         # Apply phase shift
-        for k in dataset.get_phase_keys():
-            dataset[k] = phi_coeff[inverse] * (dataset[k] + phi_shift[inverse])
-        dataset.canonicalize_phases(inplace=True)
+        for k in self.get_phase_keys():
+            self[k] = phi_coeff[inverse] * (self[k] + phi_shift[inverse])
+        self.canonicalize_phases(inplace=True)
         
-        return dataset
-            
+        return self
+
+    def expand_to_p1(self):
+        """
+        Generates all symmetrically equivalent reflections. The spacegroup 
+        symmetry is set to P1.
+        
+        Returns
+        -------
+        DataSet
+        """
+        if not self.merged:
+            raise ValueError("This function is only applicable for merged DataSets")
+
+        groupops = self.spacegroup.operations()
+        p1 = rs.concat([ self.apply_symop(op) for op in groupops ])
+        p1.spacegroup = gemmi.SpaceGroup(1)
+        p1.reset_index(inplace=True)
+        p1.drop_duplicates(subset=["H", "K", "L"], inplace=True)
+        p1.set_index(["H", "K", "L"], inplace=True)
+        return p1
+
+    def expand_anomalous(self):
+        """
+        Expands data by applying Friedel operator (-x, -y, -z). The
+        necessary phase shifts are made for columns of complex
+        dtypes or PhaseDtypes. 
+        
+        Returns
+        -------
+        DataSet
+        """
+        friedel = self.apply_symop("-x,-y,-z")
+        return self.append(friedel)
+    
+    @inplace
     def canonicalize_phases(self, inplace=False):
         """
         Canonicalize columns with phase data to fall in the interval between
@@ -956,12 +1124,49 @@ class DataSet(pd.DataFrame):
         -------
         DataSet
         """
-        if inplace:
-            dataset = self
-        else:
-            dataset = self.copy()
+        for k in self.get_phase_keys():
+            self[k] = utils.canonicalize_phases(self[k])
 
-        for k in dataset.get_phase_keys():
-            dataset[k] = utils.canonicalize_phases(dataset[k])
+        return self
 
-        return dataset
+    def to_reciprocalgrid(self, key, gridsize, friedels_law=True):
+        """
+        Set up reciprocal grid with values from column, ``key``, indexed by
+        Miller indices. A 3D numpy array will be initialized with the
+        specified gridsize. Any missing Miller indices are initialized to
+        zero.
+
+        Notes
+        -----
+        - The data being arranged on a reciprocal grid must be compatible
+          with a numpy datatype.
+        - If the data is complex and ``friedels_law=True``, the complex
+          conjugate will be used to populate Friedel pairs. The only scenario
+          in which to use ``friedels_law=False`` would be if Friedel pairs
+          have already been stacked to go from two-column to one-column
+          anomalous data. 
+        
+        Parameters
+        ----------
+        key : str
+            Column label for value to arrange on reciprocal grid
+        gridsize : array-like (len==3)
+            Dimensions for 3D reciprocal grid.
+        friedels_law : bool
+            Whether to expand data from P1 ASU to P1 unit cell when constructing
+            reciprocal grid. If False, Friedel pairs must be handled by
+            the user
+
+        Returns
+        -------
+        numpy.ndarray
+        """
+        p1 = self.expand_to_p1()
+        if friedels_law:
+            p1 = p1.expand_anomalous()
+        p1.sort_index(inplace=True)
+        data = p1[key].to_numpy()
+        H = p1.get_hkls()
+        grid = np.zeros(gridsize, dtype=data.dtype)
+        grid[H[:, 0], H[:, 1], H[:, 2]] = data
+        return grid
