@@ -38,6 +38,48 @@ def test_constructor_dataset(data_fmodel, spacegroup, cell, merged):
         assert result.cell == data_fmodel.cell
 
 
+@pytest.mark.parametrize("spacegroup", [None, 1, 5, 19, "P 21 21 21", "R 3:h", "R 3:blah", 1.2])
+def test_spacegroup(data_fmodel, spacegroup):
+    if spacegroup != 1.2 and spacegroup != "R 3:blah":
+        data_fmodel.spacegroup = spacegroup
+        if isinstance(spacegroup, (str, int)):
+            assert data_fmodel.spacegroup.xhm() == gemmi.SpaceGroup(spacegroup).xhm()
+        else:
+            assert data_fmodel.spacegroup == spacegroup
+    else:
+        with pytest.raises(ValueError):
+            data_fmodel.spacegroup = spacegroup
+
+
+@pytest.mark.parametrize("cell", [None,
+                                  gemmi.UnitCell(10, 20, 30, 40, 50, 60),
+                                  [10, 20, 30, 40, 50, 60],
+                                  (10, 20, 30, 40, 50, 60),
+                                  np.array([10, 20, 30, 40, 50, 60]),
+                                  [10, 20, 30],
+                                  []
+])
+def test_cell(data_fmodel, cell):
+    if cell is None:
+        data_fmodel.cell = cell
+        assert data_fmodel.cell is None
+    elif not isinstance(cell, np.ndarray) and (cell == [] or cell == [10, 20, 30]):
+        with pytest.raises(ValueError):
+            data_fmodel.cell = cell
+    else:
+        data_fmodel.cell = cell
+        if isinstance(cell, gemmi.UnitCell):
+            expected = cell
+        else:
+            expected = gemmi.UnitCell(*cell)
+        assert expected.a == data_fmodel.cell.a
+        assert expected.b == data_fmodel.cell.b
+        assert expected.c == data_fmodel.cell.c
+        assert expected.alpha == data_fmodel.cell.alpha
+        assert expected.beta == data_fmodel.cell.beta
+        assert expected.gamma == data_fmodel.cell.gamma
+
+
 @pytest.mark.parametrize("spacegroup", [None, gemmi.SpaceGroup(1)])
 @pytest.mark.parametrize("cell", [None, gemmi.UnitCell(1, 1, 1, 90, 90, 90)])
 def test_constructor_gemmi(data_gemmi, spacegroup, cell):
@@ -55,7 +97,28 @@ def test_constructor_gemmi(data_gemmi, spacegroup, cell):
     else:
         assert result.cell == data_gemmi.cell
 
+def test_to_structurefactor(data_fmodel):
+    """Test DataSet.to_structurefactor()"""
+    result = data_fmodel.to_structurefactor("FMODEL", "PHIFMODEL")
+    sfamps = data_fmodel["FMODEL"].to_numpy()
+    phases = data_fmodel["PHIFMODEL"].to_numpy()
+    expected = sfamps*np.exp(1j*np.deg2rad(phases))
+    assert isinstance(result, rs.DataSeries)
+    assert np.allclose(result.to_numpy(), expected)
 
+
+def test_from_structurefactor(data_fmodel):
+    """Test DataSet.from_structurefactor()"""
+    data_fmodel["sf"] = data_fmodel.to_structurefactor("FMODEL", "PHIFMODEL")
+    f, phi = data_fmodel.from_structurefactor("sf")
+    assert isinstance(f, rs.DataSeries)
+    assert isinstance(phi, rs.DataSeries)
+    assert isinstance(f.dtype, rs.StructureFactorAmplitudeDtype)
+    assert isinstance(phi.dtype, rs.PhaseDtype)
+    assert np.allclose(f.to_numpy(), data_fmodel["FMODEL"].to_numpy())
+    assert np.allclose(np.sin(np.deg2rad(phi.to_numpy())), np.sin(np.deg2rad(data_fmodel["PHIFMODEL"].to_numpy())), atol=1e-6)
+    assert np.allclose(np.cos(np.deg2rad(phi.to_numpy())), np.cos(np.deg2rad(data_fmodel["PHIFMODEL"].to_numpy())), atol=1e-6)    
+    
 @pytest.mark.parametrize("skip_problem_mtztypes", [True, False])
 def test_to_gemmi_roundtrip(data_gemmi, skip_problem_mtztypes):
     """Test DataSet.to_gemmi() and DataSet.from_gemmi() roundtrip"""
@@ -135,6 +198,27 @@ def test_label_centrics(data_fmodel, inplace, no_sg):
         assert "CENTRIC" in result
         assert result["CENTRIC"].dtype.name == "bool"
 
+@pytest.mark.parametrize("inplace", [True, False])
+@pytest.mark.parametrize("no_sg", [True, False])
+def test_label_absences(data_fmodel, inplace, no_sg):
+    """Test DataSet.label_absences()"""
+    if no_sg:
+        data_fmodel.spacegroup = None
+        with pytest.raises(ValueError):
+            result = data_fmodel.label_absences(inplace=inplace)
+    else:
+        result = data_fmodel.label_absences(inplace=inplace)
+
+        # Test inplace
+        if inplace:
+            assert id(result) == id(data_fmodel)
+        else:
+            assert id(result) != id(data_fmodel)
+            
+        # Test centric column
+        assert "ABSENT" in result
+        assert result["ABSENT"].dtype.name == "bool"
+
 
 @pytest.mark.parametrize("inplace", [True, False])
 @pytest.mark.parametrize("index", [True, False])
@@ -164,7 +248,7 @@ def test_infer_mtz_dtypes_rangeindex(data_merged, inplace, index):
     else:
         assert id(result) != id(temp)
 
-    
+
 @pytest.mark.parametrize("inplace", [True, False])
 @pytest.mark.parametrize("cell", [
     gemmi.UnitCell(10., 20., 30., 90., 90., 90.,),
@@ -190,6 +274,37 @@ def test_compute_dHKL(dataset_hkl, inplace, cell):
     assert np.allclose(result["dHKL"].to_numpy(), expected)
     assert isinstance(result["dHKL"].dtype, rs.MTZRealDtype)
 
+@pytest.mark.parametrize("inplace", [True, False])
+@pytest.mark.parametrize("include_centering", [True, False])
+@pytest.mark.parametrize("spacegroup", [
+    gemmi.SpaceGroup(19),
+    gemmi.SpaceGroup(4),
+])
+def test_compute_multiplicity(dataset_hkl, inplace, include_centering, spacegroup):
+    """Test DataSet.compute_multiplicity()"""
+    dataset_hkl.spacegroup = spacegroup 
+    result = dataset_hkl.compute_multiplicity(inplace=inplace, include_centering=include_centering)
+
+    # Test inplace
+    if inplace:
+        assert id(result) == id(dataset_hkl)
+    else:
+        assert id(result) != id(dataset_hkl)
+
+    # Compare to gemmi result
+    expected = np.zeros(len(result), dtype=np.int32)
+    ops = spacegroup.operations()
+    if include_centering:
+        for i, h in enumerate(result.get_hkls()):
+            expected[i] = ops.epsilon_factor(h)
+    else:
+        for i, h in enumerate(result.get_hkls()):
+            expected[i] = ops.epsilon_factor_without_centering(h)
+
+    assert np.all(result["EPSILON"].to_numpy() == expected)
+    assert isinstance(result["EPSILON"].dtype, rs.MTZIntDtype)
+
+
 @pytest.mark.parametrize("bins", [5, 10, 20, 50])
 @pytest.mark.parametrize("inplace", [True, False])
 @pytest.mark.parametrize("return_labels", [True, False])
